from __future__ import nested_scopes
import sys
from functional import dict, Ref, curry
import string
import time
import types

class FuncObject:
    def __copy__(self):
        pr = PropertyRecord(forCopy = 1)
        pr.__dict__.update(self.__dict__)
        return pr

    def set(self, name, value):
        ob = self.__copy__()
        ob.__dict__[name] = value
        return ob

    def incr(self, name, amt = 1):
        ob = self.__copy__()
        ob.__dict__[name] += amt
        return ob
        
    def decr(self, name, amt = 1):
        ob = self.__copy__()
        ob.__dict__[name] -= amt
        return ob

    def flip(self, name):
        ob = self.__copy__()
        ob.__dict__[name] = not getattr(self, name)
        return ob
        

class PropertyRecord(FuncObject):
    def __init__(self, forCopy = 0):
        if forCopy:
            return
        self.__dict__['b'] = 0
        self.__dict__['em'] = 0
        self.__dict__['i'] = 0
        self.__dict__['tt'] = 0
        self.__dict__['u'] = 0
        self.__dict__['size'] = 0
        self.__dict__['sizeDepth'] = 0
        self.__dict__['s'] = 0
        self.__dict__['pl'] = 0
        self.__dict__['color'] = 'default'
        self.__dict__['colorDepth'] = 0
        self.__dict__['wipe'] = 0
        self.__dict__['index'] = 0
        self.__dict__['tag'] = ""
        self.__dict__['endIndex'] = -1

    def setColor(self, c):
        ob = self.__copy__()
        ob.__dict__['color'] = c
        ob.__dict__['colorDepth'] = 0
        return ob

    def setSize(self, s):
        ob = self.__copy__()
        ob.__dict__['size'] = s
        ob.__dict__['sizeDepth'] = 0
        return ob

    def __str__(self):
        return repr(self.__dict__)

    __repr__ = __str__

Whitespace = (' ', '\t', '\n', '\r')
        
Attributes = map(intern, ["B", "EM", "I", "PL", "S", "TT", "U", "1",
              "2", "3", "4", "5", "6", "7", "8", "9",
              "r", "g", "b", "c", "y", "m", "k", "w"])

Tails = map(lambda x:intern("</%s>" % x), Attributes)

Tags = map(lambda x:intern("<%s>" % x), Attributes)

AllTags = Tags + Tails

TagForAttr = dict(zip(Attributes, Tags))

StartNames = dict(zip(map(lambda x:intern("START_%s" % x), Attributes), Tags))
EndNames = dict(zip(map(lambda x:intern("END_%s" % x), Attributes), Tags))
globals().update(StartNames)
globals().update(EndNames)

def parse(chunk):
    start = 0
    pieces = []
    lastntx = -2
    while 1:
        ntx = chunk.find('<', start)
        #print start, ntx
        if lastntx == ntx:
            raise "Loop"
        lastntx = ntx
        if ntx < 0:
            piece = chunk[start:]
            if piece:
                pieces.append(piece)
            break
        else:
            text = chunk[start:ntx]
            if text:
                pieces.append(text)
            start = ntx
            ntx = chunk.find('>', ntx)
            tag = chunk[start:ntx + 1]
            pieces.append(tag)
            start = ntx + 1
    #print "--", len(pieces)
    return pieces

SIZE = 0
COLOR = 1
STYLE = 2
NESTABLE_STYLE = 4

class Node:
    def __init__(self, tag):
        self.tag = tag
        self.children = []
        self.parent = None
        self.category = (tag in Colors and COLOR) or (tag in NestableStyles and NESTABLE_STYLE) or \
                        (tag in Styles and STYLE)

    def add(self, kid):
        self.children.append(kid)
        try:
            kid.parent = self
        except:
            pass

    def __str__(self):
        return "<Node %s with children: %s>" % (self.tag, self.children)

    def __repr__(self):
        return "<Node %s at %d>" % (self.tag, id(self))

    def pprint(self, depth = '', file = sys.stdout):
        print >> file, depth, repr(self)
        for kid in self.children:
            try:
                kid.pprint(depth + '  ', file = file)
            except:
                print >> file, depth + '   "' + repr(kid) + '"'

    def eat(self, node):
        self.children.extend(node.children)
        if node.parent:
            node.parent.children.remove(node)
        for k in node.children:
            try:
                k.parent = self
            except:
                pass

    def dissolve(self):
        if not self.parent:
            raise "No parent!"
        oldlen = len(self.parent.children)
        ntx = self.parent.children.index(self)
        #print self.parent.children
        self.parent.children[ntx:ntx + 1] = self.children
        assert len(self.parent.children) == oldlen + len(self.children) -1, self.parent.children
        for kid in self.children:
            try:
                kid.parent = self.parent
            except:
                pass
        
    def remove(self, node):
        self.children.remove(node)
        node.parent = None

    def splitAt(self, ntx):
        n = Node(self.tag)
        n.parent = self.parent
        n.children = self.children[ntx:]
        del self.children[ntx:]
        return n

    def insert(self, index, node):
        self.children.insert(index, node)
        try:
            node.parent = self
        except:
            pass

    def tolist(self):
        lst = []
        if self.tag:
            lst.append(self.tag)
        for kid in self.children:
            try:
                lst.extend(kid.tolist())
            except:
                lst.append(kid)
        if self.tag:
            lst.append(self.tag[0] + '/' + self.tag[1:])
        return lst
                
def makeDom(pieces):
    root = Node(None) #The root context/document
    current = root
    for piece in pieces:
        if piece is None or not piece:
            continue
        else:
            #print piece
            if piece[0] == '<':
                if piece[1] == '/':
                    current = current.parent
                else:
                    n = Node(piece)
                    current.add(n)
                    current = n
            else:
                current.add(piece)
    return root

def simplifyWS(node, pl = None):
    if DEADLINE - time.time() < 6:
        #Save 20 seconds for writing out...
        print >> sys.stderr, "Out of time, start writing!"
        return
    lastKid = None
    changes = 0
    lenChildren = len(node.children)
##    if lenChildren > 2 and isPlain(node):
##        plinserts = []
##        for i in range(lenChildren)[1:-1]:
##            if type(node.children[i]) == types.StringType:
##                prevKid = node.children[i-1]                
##                if isinstance(prevKid, Node) and not prevKid.tag == "<EM>":
##                    nextKid = node.children[i + 1]
##                    if isinstance(nextKid, Node) and not nextKid.tag == "<EM>":
##                        plinserts.append(i)
##        plinserts.reverse()
##        for i in plinserts:
##            n = Node("<PL>")
##            n.add(node.children[i])
##            node.children[i] = n
##            n.parent = node
##        print >> stackfile, "**Added <PL>s"
##        node.pprint(file = stackfile)
    string = 0
    for kid in node.children:
        if type(kid) == types.StringType:
            string = 1
        else:
            #print kid
            if not kid.children:
                #print >> stackfile, "No children, removing"
                node.remove(kid)
                changes = 1
                break
            if kid.tag == node.tag:
                if kid.tag in ColorHeads or kid.tag in SizeHeads or kid.tag in  NestableStyles or\
                   kid.tag == '<U>' and uLevels(node) >= 3:
                    kid.dissolve()
                    changes = 1
                    break
            if kid.tag == "<PL>":
                pl = kid
                continue
            if lastKid:
                #print "Last:", lastKid
                if lastKid.tag == kid.tag and not (pl or string):
                    #print >> stackfile, "Eating:", lastKid, kid
                    lastKid.eat(kid)
                    #node.pprint(file = stackfile)
                    changes = 1
                    break
                else:
                    k1 = lastKid.children[0]
                    k2 = kid.children[0]
                    oneKidEach = 0
                    if len(lastKid.children) == len(kid.children) == 1:
                        oneKidEach = 1
                    k1IsNode = isinstance(k1, Node)
                    k2IsNode = isinstance(k2, Node)
                    bothKsAreNodes = k1IsNode and k2IsNode
                    if bothKsAreNodes:
                        if k1.tag == k2.tag:
                            if not k1.tag in ColorHeads:
                                liftAndMerge(node, lastKid, kid, k1, k2, not oneKidEach)
                                changes = 1
                            break
                        #print k1, k2
                    if k1IsNode and k1.tag == kid.tag and len(k1.children) <= 1 and not (pl or string):
                        promote(node, kid, lastKid, k1, 'insert')
                        changes = 1
                        break
                    elif k2IsNode and k2.tag == lastKid.tag and len(k2.children) <= 1 and not (pl or string):
                        promote(node, lastKid, kid, k2, 'append')
                        changes = 1
                        break
                    #This does the simple color merge stuff..
                    elif lastKid.category in (COLOR, SIZE) and kid.category in (COLOR, SIZE) and not (pl or string):
                        nest(node, lastKid, kid)
                        changes = 1
                        break
            lastKid = kid
            pl = None
            string = 0
    if changes:
        simplifyWS(node, pl)
    else:
        for kid in node.children:
            if type(kid) != types.StringType:
                simplifyWS(kid)

def uLevels(node):
    u = 0
    while 1:
        if node is None or node.tag == '<PL>':
            break
        if node.tag == "<U>":
            u += 1
        node = node.parent
        
def isPlain(node):
    if node.parent is None:
        return 1
    elif node.tag == "<PL>":
        return 1
    else:
        return 0

def hasColor(node):
    while 1:
        if node.parent is None:
            return 0
        elif node.tag in ColorHeads:
            return 1
        else:
            node = node.parent
    

def promote(node, gainingNode, nodeToMove, nodeToDissolve, insertOrAppend):
    """
    B/Z
    C/B/X
    =>
    B/Z C/X
    """
    #node.pprint(file = stackfile)
    nodeToDissolve.dissolve()
    node.remove(nodeToMove)
    if insertOrAppend == 'insert':
        gainingNode.insert(0, nodeToMove)
    else:
        gainingNode.add(nodeToMove)
    #print >> stackfile, "--promote--"
    #node.pprint(file = stackfile)

def liftAndMerge(node, leftParent, rightParent, leftKid, rightKid, split):
    """
    A/B C
    Z/B
    =>
    B/A Z
    C
    """
    #node.pprint(file = stackfile)
    newParent = Node(leftKid.tag)
    leftKid.dissolve()
    rightKid.dissolve()
    lntx = node.children.index(leftParent)
    rntx = node.children.index(rightParent)
    between = node.children[lntx + 1:rntx]
    del node.children[lntx:rntx + 1]
    node.insert(lntx, newParent)
    newParent.add(leftParent)
    for thing in between:
        if isinstance(thing, Node):
            newParent.add(thing)
        else:
            newPl = Node("<PL>")
            newPl.add(thing)
            newParent.add(newPl)
    newParent.add(rightParent)
    if split:
        newRoot = rightParent.splitAt(1)
        node.add(newRoot)
    #print >> stackfile, "++++ Merge/split +++++++"
    #node.pprint(file = stackfile)
                            

def nest(node, gaining, moving):
    """
    A/B C
    =>
    A/B/C
    """
    #node.pprint(file = stackfile)
    node.remove(moving)
    gaining.add(moving)
    #print >> stackfile, "== Nest =="
    #node.pprint(file = stackfile)
    

                    

def finalOptimize(pieces, history):
    #For this piece we need a more sophisticated data structure!
    dom = makeDom(pieces)
##    if stackfile:
##        dom.pprint(file = stackfile)
##    print >> stackfile, "BEGIN DOM OPTIMIZATIONS"
    simplifyWS(dom)
    #nestColor(dom)
    return dom.tolist()


def unparse(file, pieces):
    #print "Unparse:", len(pieces)
    for piece in pieces:
        if not piece is None:
            file.write(piece)

ColorHeads = map(lambda x:"<%s>" % x, ["r", "g", "b", "c", "m", "y", "k"])
ColorTails = map(lambda x:"</%s>" % x, ["r", "g", "b", "c", "m", "y", "k"])
Colors = ColorHeads + ColorTails

ColorValues = dict(map(lambda x: (x, x[-2]), Colors))

SizeHeads = map(lambda x:"<%s>" % x, ["1", "2", "3", "4", "5", "6", "7", "8", "9"])
SizeTails = map(lambda x:"</%s>" % x, ["1", "2", "3", "4", "5", "6", "7", "8", "9"])
Sizes = SizeHeads + SizeTails

NestableStyles = map(lambda x:"<%s>" % x, ["B", "I", "S", "U", "PL", "TT"]) + \
        map(lambda x:"</%s>" % x, ["B", "I", "S", "U", "PL", "TT"])

Styles = NestableStyles + ['<EM>', '</EM>']
SizeValues = dict(map(lambda x: (x, x[-2]), Sizes))

Ems = ['<EM>', '</EM>']

stackfile = None

def optimize(pieces):
    #This function uses nested scopes and Refs to
    #allow functions defined in 'optimize' to share
    #these variables which are relevant to all of
    #them.

    #optimize works with a context stack, a context history,
    #and a few other state variables that allow it
    #to examine patterns in the document.
    
    context = Ref(PropertyRecord())
    ctx_stack = [context.val]
    lastWhite = 0
    allWhite = 0
    shouldPush = Ref(0)
    history = [None] * len(pieces)
    maxI = Ref(len(pieces))
    lastTextIndex = Ref(-1)
    textIndices = []
    speculations = []
    
    def wipe():
##        print "Wipe:", piece
        pieces[i] = None
        
    def push(ctx):
        ctx_stack.append(ctx)
        context.val = ctx
        ctx.__dict__['index'] = i
        ctx.__dict__['tag'] = pieces[i]
        shouldPush.val = 0

    def pushwipe(ctx, shouldWipe):
        push(ctx)
        if shouldWipe:
            wipe()
            ctx.__dict__['wipe'] = 1

    def pop():
        context.val.__dict__['endIndex'] = i
        ctx_stack.pop()
        context.val = ctx_stack[-1]

    def handleStart(piece):
        shouldPush.val = 1
        if piece == "<TT>":
            push(context.val.incr("tt"))
        elif piece == "<PL>":
            push(PropertyRecord())
        elif piece == "<U>":
            pushwipe(context.val.incr("u"), context.val.u >= 3)
        elif piece == "<S>":
            pushwipe(context.val.incr("s"), context.val.s)
        elif piece == "<B>":
            pushwipe(context.val.incr("b"), context.val.b)
        elif piece == "<I>":
            pushwipe(context.val.incr("i"), context.val.i)
        elif piece == "<PL>":
            pushwipe(context.val.incr("pl"), context.val.pl)
        elif piece == "<EM>":
            pushwipe(context.val.flip("em"), context.val.s)
##        elif not context.val.u and piece in Colors:
##            wipe()
        elif piece in Colors:
            val = ColorValues[piece]
            same = (val == context.val.color)
            if same:
                pushwipe(context.val.incr("colorDepth"), same)
            else:
                ctx = context.val.setColor(val) 
                push(ctx)
        elif piece in Sizes:
            val = SizeValues[piece]
            same = context.val.size == val
            if same:
                pushwipe(context.val.incr("sizeDepth"), same)
            else:
                push(context.val.setSize(val))

##        if piece in FloatableStyles:
##            if len(floatStyles):
##                if floatStyles[-1].endIndex < lastTextIndex.val:
##                    #We've encountered text since the last floatable style ended,
##                    #now check whether it matters for this style.
##                #print "Trying styles:", floatStyles
##                tryNest(piece, floatStyles)
                
        if shouldPush.val:
            push(context.val)
##        if stackfile:
##            print >> stackfile, "After start", piece, ": Depth", len(ctx_stack), "ctx", context.val

    def tryNest(tag, stack):
        try:
            z = -1
            while 1:                
                test = stack[z]
                if test.endIndex < 0:
                    #There's still an open tag, no nesting possible yet.
##                    print "Still open, can't nest."
                    return
##                print test.tag, tag
                if test.tag == tag:
##                    print "Nested!"
                    
                    #throw away the matching item and anything 'below'....                    
                    del stack[z:]

                    #Clear the start tag, since it's become redundant                                
                    pieces[context.val.index] = None
                    
                    #Clear the end tag for the thing that's now being nested.
                    #It will now be ended by the tag that would have ended this
                    #thing. Note that the context between test.index and
                    #here will be wrong wrt. to this tag.
                    pieces[test.endIndex] = None
                z -= 1
        except IndexError:
            pass

    def eliminateWSFormatting():
        #Check if everything between now and the last tag
        #is whitespace or none, and if so, wipe the tag start
        #and end.
        start = ctx_stack[-1].index
        ntx = start + 1
        while ntx < i:
            p = pieces[ntx]
            if p is None or not p.strip():
                ntx += 1
            else:
                return
        #If we get here, there's nothing but whitespace and
        #deleted tags between here and there, so the tag is
        #deletable.
##        print "eliminateWSFormatting killing:", (pieces[start], pieces[i])
        pieces[start] = None
        pieces[i] = None

    def emEliminate():
        #Search context stack for em changing more than once,
        #If there's no text between the two we can eliminate
        #both.
        cnt = len(ctx_stack)-2
        while cnt >= 0:
            ctx = ctx_stack[cnt]
            cnt -= 1
            if ctx.tag == "<EM>":
                ntx = start = ctx.index
                ntx += 1
                while ntx < context.val.index: #search up to the most recent <EM>.
                    p = pieces[ntx]
                    if p is None or p in AllTags or not p.strip():
                        ntx += 1
                    else:
                        return
                #If we get here there's no text, safe to eliminate both.
                pieces[start] = None
                pieces[context.val.index] = None
                pieces[i] = None
                #instruct the optimizer to eat the next matching /EM tag...
                ctx.__dict__['wipe'] = 1
                
    
    def handleEnd(piece):
        if context.val.wipe:
            wipe()
        else:
            if piece == "</U>":
                if allWhite: eliminateWSFormatting()
            elif piece == "</S>":
                if allWhite: eliminateWSFormatting()
            elif piece == "</B>":
                if allWhite: eliminateWSFormatting()
            elif piece == "</I>":
                if allWhite: eliminateWSFormatting()
            elif piece in Colors:
                if allWhite and not context.val.u: eliminateWSFormatting()
            elif piece == "</EM>":
                if allWhite:
                    eliminateWSFormatting()
                else:
                    emEliminate()
            if not pieces[i] is None:
                if lastTextIndex.val < context.val.index:
##                    print "No text, so wipe"
                    #If there's been no intervening text, there's no effect!
                    wipe()
                    pieces[context.val.index] = None
                elif piece in Colors:
                    tag = checkColorOverlap()
        pop()
##        if stackfile:
##            print >> stackfile, "After end", piece, ": Depth", len(ctx_stack), "ctx", context.val

    def checkColorOverlap():
        #Check if this color range contains any text, or just other colors which
        #contain text.
        #Expensive, speedwise.
        ntx = context.val.index + 1
        headQual = 0
#        print "Checking overlap from", ntx
        while 1:
            p  = pieces[ntx]
#            print ntx, p
            if p is None:
#                print "Found None"
                ntx += 1
            elif p in ColorHeads:
                headQual  = 1
                break                
            elif not p in AllTags:
#                print "Found text"
                return piece
            else:
                ntx += 1
        if not headQual:
#            print "Searched to no avail."
            return piece
#        print "HeadQual!"
        ntx2 = i -1
        while ntx2 > ntx:
            if ntx in textIndices:
                return piece
            else:
                if pieces[ntx] in ColorTails:
                    #We've hit no text before or after, so
                    #our current color is irrelevant.
##                    print "ColorOverlap!"
                    pieces[context.val.index] = None
                    pieces[i] = None
                    return pieces[ntx]
                ntx -= 1
    i = 0
    while i < maxI.val:
        piece = pieces[i]
        if piece[0] == "<":
            if piece[1] != "/":
                handleStart(piece)
                history[i] = context.val
            else:
                history[i] = context.val
                handleEnd(piece)
        else:
            lastTextIndex.val = i
            textIndices.append(i)
            if not context.val.tt:
                pieces[i], lastWhite = strip_whitespace(piece, lastWhite)
            else:
                #In tt, whitespace significant.
                lastWhite = 0
            allWhite = not pieces[i] or not pieces[i].strip()
        i += 1
    i = 0

    return pieces, history


def strip_whitespace(s, lastWhite):
    build = []
    for char in s:
        if char in Whitespace:
            if lastWhite:
                continue
            else:
                build.append(COMPRESSION_CHAR)
                lastWhite = 1
        else:
            build.append(char)
            lastWhite = 0
    res = string.join(build, "")
##    diff = len(s) - len(res)
##    if diff:
##        print "Stripped", diff
    return res, lastWhite

def main(input, output):
    start = time.time()
    data = input.read()
    unparse(output, finalOptimize(*optimize(parse(data))))
    #print "Finished in", time.time() - start

def checkInput(input, expected):
    inp = cStringIO.StringIO(input)
    outp = cStringIO.StringIO()
    main(inp, outp)
    if outp.getvalue() != expected:
        print "Input:\t\t", inp.getvalue()
        print "Should be:\t", expected
        print "Was:\t\t", outp.getvalue()
        print "===FAILED==="
    else:
        print "PASS"


if __name__ == '__main__':
    if sys.argv[1] == 'test':
        import os, stat, cStringIO
        COMPRESSION_CHAR = ' '
        #Test whitespace compression
        name = "compression"
        stackfile = open(name + ".stack", 'wb')
        checkInput("   <B> foo   bar </B>", " <B>foo bar </B>")

        #Redundancy elimination
        name = "redundancy"
        stackfile = open(name + ".stack", 'wb')
        checkInput("<B><I><B>foo</B></I></B>", "<B><I>foo</I></B>")
        
        #Test color nesting
        name = "colornest"
        stackfile = open(name + ".stack", 'wb')
        checkInput("<r>aaa</r><g>bbb</g><b>ccc</b><g>ddd</g><r>eee</r>",
                   "<r>aaa<g>bbb<b>ccc</b>ddd</g>eee</r>")

        #Test color nesting gotcha
        name = "colorgotcha"
        stackfile = open(name + ".stack", 'wb')
        checkInput("<r>aaa<g>bbb</g></r><g><r>ccc</r></g>",
                   "<r>aaa<g>bbb</g>ccc</r>")

        #overlap inversion
        stackfile = open("overlap.stack", 'wb')
        checkInput("<B><I>bla bla</I></B><TT><I><B>foo bar</B></I> truc</TT>",
                    "<B><I>bla bla<TT>foo bar</TT></I></B><TT> truc</TT>")
        #PL shortcut
        stackfile = open("pl.stack", 'wb')
        checkInput("<I><S><TT>foo</TT></S></I> bar <TT><S><I> gee</I></S></TT>",
                    "<I><S><TT>foo<PL> bar </PL> gee</TT></S></I>")

        #whitespace simplification
        stackfile = open("whitesimple.stack", 'wb')
        checkInput("<B><I><r>bla bla </r> </I> </B> <r><I> barfoo</I></r>",
                    "<I><r><B>bla bla </B>barfoo</r></I>")

        
        if raw_input("Continue?") == 'n':
            raise SystemExit
        COMPRESSION_CHAR = '\n'
        files = filter(lambda x:x.endswith('input'), os.listdir('.'))
        for f in files:
            outfile = f[:f.find(".") + 1] + "output"
            stackfile = open(f[:f.find(".") + 1] + "stack", 'wb')
            input = open(f, 'rb')
            contents = input.read()
            output = open(outfile, 'wb')
            unparse(output, parse(contents))
            output.close()
            output = open(outfile, 'rb')
            c2 = output.read()
            assert contents == c2
            output.close()
            output = open(outfile, 'wb')
            input.seek(0)
            main(input, output)
            output.close()
            input.close()
            s1, s2 = (os.stat(f), os.stat(outfile))
            print s1[stat.ST_SIZE], "/", s2[stat.ST_SIZE]
        
    else:
        DEADLINE = time.time() + float(sys.argv[1])
        COMPRESSION_CHAR = ' '
        input = sys.stdin
        output = sys.stdout
        main(input, output)
        output.flush()
                
    
        
    
